@@ -63,6 +63,8 @@ typedef struct _GdmLocale {
 static GHashTable *gdm_languages_map;
 static GHashTable *gdm_territories_map;
 static GHashTable *gdm_available_locales_map;
+static GHashTable *gdm_language_count_map;
+static GHashTable *gdm_territory_count_map;
 
 static char * construct_language_name (const char *language,
                                        const char *territory,
@@ -321,8 +323,8 @@ language_name_get_codeset_details (const char  *language_name,
         g_free (old_locale);
 }
 
-static gboolean
-language_name_has_translations (const char *language_name)
+gboolean
+gdm_language_has_translations (const char *language_name)
 {
         GDir        *dir;
         char        *path;
@@ -403,6 +405,7 @@ add_locale (const char *language_name,
 #ifdef WITH_INCOMPLETE_LOCALES
         if (utf8_only) {
                 if (locale->territory_code == NULL || locale->modifier) {
+                        g_debug ("Ignoring '%s' as a locale, since it lacks territory code or modifier", name);
                         gdm_locale_free (locale);
                         return FALSE;
                 }
@@ -415,9 +418,9 @@ add_locale (const char *language_name,
                                                 locale->codeset, locale->modifier);
 
 #ifndef WITH_INCOMPLETE_LOCALES
-        if (!language_name_has_translations (locale->name) &&
-            !language_name_has_translations (locale->id) &&
-            !language_name_has_translations (locale->language_code) &&
+        if (!gdm_language_has_translations (locale->name) &&
+            !gdm_language_has_translations (locale->id) &&
+            !gdm_language_has_translations (locale->language_code) &&
             utf8_only) {
                 g_debug ("Ignoring '%s' as a locale, since it lacks translations", locale->name);
                 gdm_locale_free (locale);
@@ -612,6 +615,32 @@ collect_locales_from_locale_file (const char *locale_file)
         fclose (langlist);
 }
 
+static void
+count_languages_and_territories (void)
+{
+	gpointer value;
+	GHashTableIter iter;
+	gint count;
+
+	gdm_language_count_map = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+	gdm_territory_count_map = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+
+        g_hash_table_iter_init (&iter, gdm_available_locales_map);
+        while (g_hash_table_iter_next (&iter, NULL, &value)) {
+                GdmLocale *locale;
+
+                locale = (GdmLocale *) value;
+
+		count = GPOINTER_TO_INT (g_hash_table_lookup (gdm_language_count_map, locale->language_code));
+		count++;
+		g_hash_table_insert (gdm_language_count_map, g_strdup (locale->language_code), GINT_TO_POINTER (count));
+
+		count = GPOINTER_TO_INT (g_hash_table_lookup (gdm_territory_count_map, locale->territory_code));
+		count++;
+		g_hash_table_insert (gdm_territory_count_map, g_strdup (locale->territory_code), GINT_TO_POINTER (count));
+        }
+}
+
 static void
 collect_locales (void)
 {
@@ -626,10 +655,45 @@ collect_locales (void)
                            "guessing possible locales from available translations, "
                            "but list may be incomplete!");
 #endif
-
-                collect_locales_from_directory ();
         }
+
+        collect_locales_from_directory ();
+
         collect_locales_from_locale_file (ALIASES_FILE);
+
+	count_languages_and_territories ();
+}
+
+static gint
+get_language_count (const char *language)
+{
+        if (gdm_language_count_map == NULL) {
+                collect_locales ();
+        }
+
+	return GPOINTER_TO_INT (g_hash_table_lookup (gdm_language_count_map, language));
+}
+
+static gboolean
+is_unique_language (const char *language)
+{
+  return get_language_count (language) == 1;
+}
+
+static gint
+get_territory_count (const char *territory)
+{
+        if (gdm_territory_count_map == NULL) {
+                collect_locales ();
+        }
+
+	return GPOINTER_TO_INT (g_hash_table_lookup (gdm_territory_count_map, territory));
+}
+
+static gboolean
+is_unique_territory (const char *territory)
+{
+  return get_territory_count (territory) == 1;
 }
 
 static gboolean
@@ -1100,6 +1164,10 @@ gdm_get_language_from_name (const char *name,
 
         full_language = g_string_append (full_language, translated_language);
 
+	if (is_unique_language (language_code)) {
+		goto out;
+	}
+
         if (territory_code != NULL) {
                 translated_territory = get_translated_territory (territory_code, locale);
         }
@@ -1137,6 +1205,94 @@ out:
        return g_string_free (full_language, FALSE);
 }
 
+char *
+gdm_get_region_from_name (const char *name,
+                          const char *locale)
+{
+        GString *full_name;
+        char *language_code;
+        char *territory_code;
+        char *codeset_code;
+        char *langinfo_codeset;
+        char *translated_language;
+        char *translated_territory;
+        gboolean is_utf8 = TRUE;
+
+        g_return_val_if_fail (name != NULL, NULL);
+        g_return_val_if_fail (*name != '\0', NULL);
+
+        translated_territory = NULL;
+        translated_language = NULL;
+        langinfo_codeset = NULL;
+
+        full_name = g_string_new (NULL);
+
+        if (gdm_languages_map == NULL) {
+                languages_init ();
+        }
+
+        if (gdm_territories_map == NULL) {
+                territories_init ();
+        }
+
+        language_code = NULL;
+        territory_code = NULL;
+        codeset_code = NULL;
+
+        gdm_parse_language_name (name,
+                                 &language_code,
+                                 &territory_code,
+                                 &codeset_code,
+                                 NULL);
+
+        if (territory_code == NULL) {
+                goto out;
+        }
+
+        translated_territory = get_translated_territory (territory_code, locale);
+        g_string_append (full_name, translated_territory);
+
+	if (is_unique_territory (territory_code)) {
+		goto out;
+	}
+
+        if (language_code != NULL) {
+                translated_language = get_translated_language (language_code, locale);
+        }
+        if (translated_language != NULL) {
+                g_string_append_printf (full_name,
+                                        " (%s)",
+                                        translated_language);
+        }
+
+        language_name_get_codeset_details (name, &langinfo_codeset, &is_utf8);
+
+        if (codeset_code == NULL && langinfo_codeset != NULL) {
+            codeset_code = g_strdup (langinfo_codeset);
+        }
+
+        if (!is_utf8 && codeset_code) {
+                g_string_append_printf (full_name,
+                                        " [%s]",
+                                        codeset_code);
+        }
+
+out:
+       g_free (language_code);
+       g_free (territory_code);
+       g_free (codeset_code);
+       g_free (langinfo_codeset);
+       g_free (translated_language);
+       g_free (translated_territory);
+
+       if (full_name->len == 0) {
+               g_string_free (full_name, TRUE);
+               return NULL;
+       }
+
+       return g_string_free (full_name, FALSE);
+}
+
 char **
 gdm_get_all_language_names (void)
 {
